@@ -1,6 +1,6 @@
 /*
  *  tenace - bridge hand viewer and editor
- *  Copyright (C) 2005-2007 Christoph Berg <cb@df7cb.de>
+ *  Copyright (C) 2005-2008 Christoph Berg <cb@df7cb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -27,6 +27,7 @@
 #include "bridge.h"
 #include "file.h"
 #include "functions.h"
+#include "options.h"
 #include "solve.h"
 #include "window_bids.h"
 #include "window_board.h"
@@ -96,6 +97,7 @@ on_window_hand_delete_event            (GtkWidget       *widget,
                                         gpointer         user_data)
 {
 	gtk_main_quit ();
+	return FALSE;
 }
 
 
@@ -138,7 +140,7 @@ on_toolbutton_card_wipe_clicked        (GtkToolButton   *toolbutton,
 	board *b = CUR_BOARD;
 	board_clear(b);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS | REDRAW_TRICKS);
+	show_board(b, REDRAW_HANDS | REDRAW_TRICKS | REDRAW_PAR);
 }
 
 
@@ -149,7 +151,8 @@ on_toolbutton_card_random_clicked      (GtkToolButton   *toolbutton,
 	board *b = CUR_BOARD;
 	deal_random(b);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS | REDRAW_DD);
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 }
 
 
@@ -324,7 +327,19 @@ on_double_dummy1_activate              (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
 	board *b = CUR_BOARD;
-	show_board(b, REDRAW_HANDS | REDRAW_DD);
+	compute_dd_scores (b, 1);
+	show_board(b, REDRAW_HANDS);
+}
+
+
+void
+on_dd_always1_activate                 (GtkMenuItem     *menuitem,
+                                        gpointer         user_data)
+{
+	board *b = CUR_BOARD;
+	run_dd = gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (menuitem));
+	compute_dd_scores (b, run_dd);
+	show_board (b, REDRAW_HANDS);
 }
 
 
@@ -334,6 +349,7 @@ on_parscore1_activate                  (GtkMenuItem     *menuitem,
 {
 	board *b = CUR_BOARD;
 	parscore(b);
+	show_board (b, REDRAW_PAR);
 }
 
 
@@ -343,6 +359,7 @@ on_handbutton_par_clicked              (GtkToolButton   *toolbutton,
 {
 	board *b = CUR_BOARD;
 	parscore(b);
+	show_board (b, REDRAW_PAR);
 }
 
 
@@ -383,12 +400,16 @@ on_set_par1_activate                   (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
 	board *b = CUR_BOARD;
-	b->trumps = b->par_suit;
+	parscore (b);
+	if (b->trumps != b->par_suit || b->declarer != b->par_dec) {
+		board_rewind (b); /* incompatible change */
+		b->trumps = b->par_suit;
+		b->declarer = b->par_dec;
+		b->current_turn = seat_mod (b->declarer + 1);
+	}
 	b->level = b->par_level;
-	b->declarer = b->par_dec;
-	b->current_turn = seat_mod(b->declarer + 1);
 	b->doubled = b->par_tricks < b->par_level + 6;
-	show_board(b, REDRAW_CONTRACT | REDRAW_NAMES);
+	show_board (b, REDRAW_FULL);
 }
 
 
@@ -429,9 +450,12 @@ on_rewind_button_clicked               (GtkToolButton   *toolbutton,
                                         gpointer         user_data)
 {
 	board *b = CUR_BOARD;
-	rewind_card(b); /* extra call to show warning if there's nothing to do */
-	board_rewind(b);
-	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY | REDRAW_DD);
+	if (!rewind_card(b)) /* extra call to show warning if there's nothing to do */
+		return;
+	while (b->n_played_cards % 4 != 0)
+		rewind_card (b);
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY);
 }
 
 
@@ -440,8 +464,10 @@ on_button_back_clicked                 (GtkToolButton   *toolbutton,
                                         gpointer         user_data)
 {
 	board *b = CUR_BOARD;
-	rewind_card(b);
-	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY | REDRAW_DD);
+	if (!rewind_card(b))
+		return;
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY);
 }
 
 
@@ -450,8 +476,10 @@ on_button_next_clicked                 (GtkToolButton   *toolbutton,
                                         gpointer         user_data)
 {
 	board *b = CUR_BOARD;
-	next_card(b);
-	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY | REDRAW_DD);
+	if (!next_card(b))
+		return;
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY);
 }
 
 
@@ -460,43 +488,50 @@ on_button_fast_forward_clicked         (GtkToolButton   *toolbutton,
                                         gpointer         user_data)
 {
 	board *b = CUR_BOARD;
-	next_card (b); /* extra call to show warning if there's nothing to do */
-	board_fast_forward(b);
-	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY | REDRAW_DD);
+	if (!next_card (b)) /* extra call to show warning if there's nothing to do */
+		return;
+	while (b->n_played_cards % 4 != 0) {
+		if (! next_card (b))
+			break;
+	}
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY);
 }
 
 
 void
-on_button_dd_toggled                   (GtkToggleToolButton *toggletoolbutton,
+on_rewind_play1_activate               (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
 	board *b = CUR_BOARD;
-	run_dd = gtk_toggle_tool_button_get_active(toggletoolbutton);
-	show_board (b, REDRAW_HANDS | REDRAW_DD);
+	if (!rewind_card (b)) /* extra call to show warning if there's nothing to do */
+		return;
+	board_rewind (b);
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY);
 }
 
 
 void
-on_pos_north_south_activate            (GtkMenuItem     *menuitem,
+on_complete_play1_activate             (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
-	pos_score_for = 0;
+	board *b = CUR_BOARD;
+	if (!next_card (b)) /* extra call to show warning if there's nothing to do */
+		return;
+	board_fast_forward (b);
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY);
 }
 
 
 void
-on_pos_declarer_activate               (GtkMenuItem     *menuitem,
+on_button_dd_clicked                   (GtkToolButton   *toolbutton,
                                         gpointer         user_data)
 {
-	pos_score_for = 1;
-}
-
-
-void
-on_pos_current_lead_activate           (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	pos_score_for = 2;
+	board *b = CUR_BOARD;
+	compute_dd_scores (b, 1);
+	show_board(b, REDRAW_HANDS);
 }
 
 
@@ -507,7 +542,7 @@ on_deal_clear_activate                 (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	board_clear(b);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS | REDRAW_PLAY);
+	show_board(b, REDRAW_HANDS | REDRAW_PLAY | REDRAW_PAR);
 }
 
 
@@ -518,7 +553,8 @@ on_deal_random_activate                (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	deal_random(b);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS | REDRAW_DD);
+	compute_dd_scores (b, run_dd);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 }
 
 
@@ -559,7 +595,7 @@ on_window_line_entry_delete_event      (GtkWidget       *widget,
                                         GdkEvent        *event,
                                         gpointer         user_data)
 {
-	PROTECT_BEGIN;
+	PROTECT_BEGIN_BOOL;
 	GtkWidget *menuitem = lookup_widget (win->window, "deal_line");
 	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menuitem), FALSE);
 	window_line_entry_delete ();
@@ -568,7 +604,6 @@ on_window_line_entry_delete_event      (GtkWidget       *widget,
 }
 
 
-
 void
 on_dealer_west1_activate               (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
@@ -608,9 +643,11 @@ on_window_card_delete_event            (GtkWidget       *widget,
                                         GdkEvent        *event,
                                         gpointer         user_data)
 {
+	PROTECT_BEGIN_BOOL;
 	GtkCheckMenuItem *menuitem = GTK_CHECK_MENU_ITEM (lookup_widget (win->window, "cards1"));
 	gtk_check_menu_item_set_active (menuitem, FALSE);
 	window_card_delete ();
+	PROTECT_END;
 	return FALSE;
 }
 
@@ -618,10 +655,12 @@ void
 on_cards1_activate                     (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
-	if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (menuitem)))
-		window_card_init ();
-	else
+	PROTECT_BEGIN;
+	if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (menuitem))) {
+		window_card_init (win->hand_display_style);
+	} else
 		window_card_delete ();
+	PROTECT_END;
 }
 
 
@@ -630,9 +669,11 @@ on_window_bids_delete_event            (GtkWidget       *widget,
                                         GdkEvent        *event,
                                         gpointer         user_data)
 {
+	PROTECT_BEGIN_BOOL;
 	GtkCheckMenuItem *menuitem = GTK_CHECK_MENU_ITEM (lookup_widget (win->window, "bids1"));
 	gtk_check_menu_item_set_active (menuitem, FALSE);
 	window_bids_delete ();
+	PROTECT_END;
 	return FALSE;
 }
 
@@ -684,7 +725,10 @@ void
 on_style_text_activate                 (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+	PROTECT_BEGIN;
 	board_window_set_style (win, HAND_DISPLAY_STYLE_TEXT);
+	window_card_set_style (HAND_DISPLAY_STYLE_TEXT);
+	PROTECT_END;
 }
 
 
@@ -692,7 +736,10 @@ void
 on_style_cards_activate                (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+	PROTECT_BEGIN;
 	board_window_set_style (win, HAND_DISPLAY_STYLE_CARDS);
+	window_card_set_style (HAND_DISPLAY_STYLE_CARDS);
+	PROTECT_END;
 }
 
 
@@ -773,69 +820,6 @@ on_deal_paste_activate                 (GtkMenuItem     *menuitem,
 }
 
 
-void
-on_ddno_one1_activate                  (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	win->show_dd_scores = seat_none;
-	show_board (win->boards[win->cur], REDRAW_HANDS);
-}
-
-
-void
-on_ddeastwest1_activate                (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	win->show_dd_scores = east_west;
-	show_board (win->boards[win->cur], REDRAW_HANDS);
-}
-
-
-void
-on_ddnorthsouth1_activate              (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	win->show_dd_scores = north_south;
-	show_board (win->boards[win->cur], REDRAW_HANDS);
-}
-
-
-void
-on_ddall1_activate                     (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	win->show_dd_scores = seat_all;
-	show_board (win->boards[win->cur], REDRAW_HANDS);
-}
-
-
-void
-on_handseastwest1_activate             (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	win->show_hands = east_west;
-	show_board (win->boards[win->cur], REDRAW_HANDS);
-}
-
-
-void
-on_handsnorthsouth1_activate           (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	win->show_hands = north_south;
-	show_board (win->boards[win->cur], REDRAW_HANDS);
-}
-
-
-void
-on_handsall1_activate                  (GtkMenuItem     *menuitem,
-                                        gpointer         user_data)
-{
-	win->show_hands = seat_all;
-	show_board (win->boards[win->cur], REDRAW_HANDS);
-}
-
-
 void
 on_ew1_activate                        (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
@@ -843,7 +827,7 @@ on_ew1_activate                        (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	flip_hands (b, east, west);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR); /* flip invalidates parscore */
 }
 
 
@@ -854,7 +838,7 @@ on_ns1_activate                        (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	flip_hands (b, north, south);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 }
 
 
@@ -865,7 +849,7 @@ on_nw1_activate                        (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	flip_hands (b, north, west);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 }
 
 
@@ -876,7 +860,7 @@ on_ne1_activate                        (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	flip_hands (b, north, east);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 }
 
 
@@ -887,7 +871,7 @@ on_sw1_activate                        (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	flip_hands (b, south, west);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 }
 
 
@@ -898,7 +882,7 @@ on_se1_activate                        (GtkMenuItem     *menuitem,
 	board *b = CUR_BOARD;
 	flip_hands (b, south, east);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 }
 
 
@@ -924,21 +908,36 @@ on_bid_undo_clicked                    (GtkToolButton   *toolbutton,
 }
 
 
+#define TRY(x) { if (!(x)) goto end; }
+
 void
 on_bid_set_contract_clicked            (GtkToolButton   *toolbutton,
                                         gpointer         user_data)
 {
+	int i;
 	board *b = CUR_BOARD;
+	seat cur_seat = seat_mod (b->dealer + b->n_bids);
+	int passes = (b->declarer - cur_seat) % 4;
 
-	board_append_bid (b, 5 * b->level + b->trumps, 0);
+	for (i = (b->n_bids + passes - 2) % 4; i < b->n_bids; i += 4)
+		if (b->bidding[i] >= 5 && DENOM (b->bidding[i]) == b->trumps) {
+			board_statusbar (_("Suit was already bid from wrong side"));
+			goto end;
+		}
+
+	for (i = 0; i < passes; i++)
+		TRY (board_append_bid (b, bid_pass, 0));
+
+	TRY (board_append_bid (b, 5 * b->level + b->trumps, 0));
 	if (b->doubled)
-		board_append_bid (b, bid_x, 0);
+		TRY (board_append_bid (b, bid_x, 0));
 	if (b->doubled == bid_xx)
-		board_append_bid (b, bid_xx, 0);
-	board_append_bid (b, bid_pass, 0);
-	board_append_bid (b, bid_pass, 0);
-	board_append_bid (b, bid_pass, 0);
+		TRY (board_append_bid (b, bid_xx, 0));
+	TRY (board_append_bid (b, bid_pass, 0));
+	TRY (board_append_bid (b, bid_pass, 0));
+	TRY (board_append_bid (b, bid_pass, 0));
 
+end:
 	show_board(b, REDRAW_BIDDING | REDRAW_BIDDING_SCROLL);
 }
 
@@ -1024,4 +1023,3 @@ on_aboutdialog1_delete_event           (GtkWidget       *widget,
 }
 
 /* -- */
-
