@@ -1,6 +1,6 @@
 /*
  *  tenace - bridge hand viewer and editor
- *  Copyright (C) 2005-2007 Christoph Berg <cb@df7cb.de>
+ *  Copyright (C) 2005-2008 Christoph Berg <cb@df7cb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,26 +15,28 @@
 
 #include <assert.h>
 #include <ctype.h>
+#include <glib.h>
+#include <gtk/gtk.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <glib.h>
-#include <gtk/gtk.h>
 
 #include "../handdisplay/hand_display.h"
 
 #include "window_board.h"
 
+#include "file.h"
 #include "functions.h"
 #include "interface.h"
+#include "options.h"
 #include "solve.h"
 #include "support.h"
+#include "window_card.h"
 #include "window_line_entry.h"
 #include "window_play.h"
 
 window_board_t *win; // FIXME static?
 int protect = 0;
-static GString *rcfile = 0;
 
 static GdkColor bidding_non_vuln = { 0, 0.8*65535, 0.8*65535, 0.8*65535 };
 static GdkColor bidding_vuln = { 0, 0.8*65535, 0, 0 };
@@ -43,7 +45,12 @@ static void
 board_menu_select (GtkWidget *menuitem, int *n)
 {
 	PROTECT_BEGIN;
+	if (!gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (menuitem)))
+		PROTECT_RETURN;
+	if (*n == win->cur)
+		PROTECT_RETURN;
 	win->cur = *n;
+	//printf ("jump to %d\n", *n);
 	assert (0 <= *n && *n < win->n_boards);
 	show_board (win->boards[*n], REDRAW_BOARD);
 	PROTECT_END;
@@ -120,7 +127,7 @@ bidding_update (window_board_t *win, board *b, int scroll)
 	if (!b->n_bids) {
 		gtk_list_store_append (win->bidding_store, &iter);
 		gtk_list_store_set (win->bidding_store, &iter,
-				2 * col, _("•"),
+				2 * col, "•",
 				2 * col + 1, _("Dealer"),
 				-1);
 		return;
@@ -156,16 +163,23 @@ void show_board (board *b, redraw_t redraw)
 	GString *str = g_string_new(NULL);
 	assert (b);
 
-	if (redraw & REDRAW_BOARD_LIST)
+	if (redraw & REDRAW_BOARD_LIST) {
 		board_window_rebuild_board_menu (win);
+		window_options_board_list_populate ();
+	}
 
 	if (redraw & (REDRAW_TITLE | REDRAW_CONTRACT)) {
 		g_string_printf(str, "Tenace - %s (%s)", b->name->str,
 			contract_string_asc (b->level, b->trumps, b->declarer, b->doubled));
 		if (win->title) {
-			g_string_append_printf (str, " - %s", win->title->str);
+			g_string_append_printf (str, " - %s", win->title);
+			if (win->subtitle)
+				g_string_append_printf (str, " - %s", win->subtitle);
+			if (win->team1 && win->team2)
+				g_string_append_printf (str, _(" - %s vs. %s"),
+						win->team1, win->team2);
 		} else if (win->filename) {
-			char *fname = win->filename->str;
+			char *fname = win->filename;
 			if (strrchr(fname, '/'))
 				fname = strrchr(fname, '/') + 1;
 			g_string_append_printf (str, " - %s", fname);
@@ -206,6 +220,8 @@ void show_board (board *b, redraw_t redraw)
 		char *vulnmenu[] = { "vuln_none", "vuln_ns", "vuln_ew", "vuln_all" };
 		w = lookup_widget(win->window, vulnmenu[2 * b->vuln[1] + b->vuln[0]]);
 		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (w), TRUE);
+
+		window_options_board_populate ();
 	}
 
 	if (redraw & REDRAW_NAMES) {
@@ -225,6 +241,8 @@ void show_board (board *b, redraw_t redraw)
 		board_set_player_name (w, south, b->dealer == south, b->vuln[0],
 				b->declarer == south,
 				b->current_turn == south, b->hand_name[3]->str);
+
+		window_options_board_populate ();
 	}
 
 	if (redraw & REDRAW_TRICKS) {
@@ -233,22 +251,14 @@ void show_board (board *b, redraw_t redraw)
 		gtk_label_set_markup((GtkLabel*) w, str->str);
 	}
 
-	if (redraw & REDRAW_DD) {
-		//if (b->par_score == -1) {
-			//w = lookup_widget(win->window, "par_label");
-			//gtk_label_set_text(GTK_LABEL(w), "");
-		//}
-
-		if (run_dd) {
-			compute_dd_scores(b);
-
-			if (b->current_dd) {
-				g_string_printf (str, "%s",
-					score_string(b->level, b->trumps, b->declarer,
-						b->doubled, b->vuln[b->declarer % 2],
-						b->current_dd->best_score, b->current_turn));
-				solve_statusbar(str->str);
-			}
+	if (redraw & REDRAW_PAR) {
+		w = lookup_widget (win->window, "par_label");
+		if (b->par_score == -1) {
+			gtk_label_set_text (GTK_LABEL (w), "");
+		} else {
+			char *par = par_label (b);
+			gtk_label_set_markup(GTK_LABEL (w), par);
+			free (par);
 		}
 	}
 
@@ -274,6 +284,9 @@ void show_board (board *b, redraw_t redraw)
 						seat_mask (i, win->show_dd_scores))
 					hand_display_set_card_score (win->handdisp[i - 1], c,
 						card_overtricks(b, c));
+				else
+					hand_display_set_card_score (win->handdisp[i - 1], c,
+						HAND_DISPLAY_NO_SCORE);
 			}
 			hand_display_set_card_score_neg (win->handdisp[i - 1],
 					i % 2 != b->declarer % 2);
@@ -305,12 +318,30 @@ void show_board (board *b, redraw_t redraw)
 	}
 }
 
+void
+recently_used_add (char *filename)
+{
+#if GTK_CHECK_VERSION (2,10,0)
+	static GtkRecentManager *recent = NULL;
+	if (! recent)
+		recent = gtk_recent_manager_get_default ();
+
+	char buf[1024];
+	snprintf (buf, sizeof (buf), "file://%s", filename);
+	gtk_recent_manager_add_item (recent, buf);
+#endif
+}
+
 /* callbacks */
 
 static void card_clicked (HandDisplay *handdisp, int card, int *seatp)
 {
+	if (card == -1)
+		return;
+
 	PROTECT_BEGIN;
 	board *b = CUR_BOARD;
+	assert (card >= 0 && card < 52);
 	//printf("Clicked: %s for %c.\n", card_string(card), "WNES"[*seatp - 1]);
 	int redraw = 0;
 
@@ -324,8 +355,10 @@ static void card_clicked (HandDisplay *handdisp, int card, int *seatp)
 		redraw = 1;
 	}
 
-	if (redraw)
-		show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_DD | REDRAW_PLAY);
+	if (redraw) {
+		compute_dd_scores (b, run_dd);
+		show_board(b, REDRAW_HANDS | REDRAW_NAMES | REDRAW_TRICKS | REDRAW_PLAY);
+	}
 
 	PROTECT_END;
 }
@@ -363,18 +396,17 @@ static void card_leave (HandDisplay *handdisp, int card, int *seatp)
 	PROTECT_END;
 }
 
+/* gets called for target widget */
 static void
-card_drag_drop (HandDisplay *handdisp, int card, int on_card, int *seatp)
+card_drag_drop (HandDisplay *handdisp, int card, int on_card, int *to_seat)
 {
 	PROTECT_BEGIN;
 	board *b = CUR_BOARD;
-	/*
-	printf("Dropped: %s for %c.\n", card_string(card), "WNES"[*seatp - 1]);
+	printf("Dropped: %s on seat %c.\n", card_string(card), "WNES"[*to_seat - 1]);
 	if (on_card >= 0)
 		printf("Dropped on: %s.\n", card_string(on_card));
-	*/
 
-	if (b->dealt_cards[card] && b->dealt_cards[card] == *seatp) /* card didn't move */
+	if (b->dealt_cards[card] && b->dealt_cards[card] == *to_seat) /* card didn't move */
 		PROTECT_RETURN;
 
 	if (b->dealt_cards[card] && !b->cards[card]) {
@@ -382,40 +414,47 @@ card_drag_drop (HandDisplay *handdisp, int card, int on_card, int *seatp)
 		PROTECT_RETURN;
 	}
 
+	seat from_seat = b->dealt_cards[card];
 	if (on_card >= 0) { /* exchange 2 cards */
 		if (b->dealt_cards[on_card] && !b->cards[on_card]) {
 			board_statusbar(_("Card is in play and cannot be exchanged"));
 			PROTECT_RETURN;
 		}
 
-		seat from_seat = b->dealt_cards[card];
-		remove_card(b, from_seat, card);
-		remove_card(b, *seatp, on_card);
-		add_card(b, from_seat, on_card);
-		add_card(b, *seatp, card);
+		remove_card(b, *to_seat, on_card);
+		if (from_seat) {
+			remove_card(b, from_seat, card);
+			add_card(b, from_seat, on_card);
+		}
+		add_card(b, *to_seat, card);
 	} else { /* move single card */
-		if (b->hand_cards[*seatp-1] == 13) {
+		if (b->hand_cards[*to_seat-1] == 13) {
 			board_statusbar(_("Hand has already 13 cards"));
 			PROTECT_RETURN;
 		}
 
-		if (b->dealt_cards[card])
-			remove_card(b, b->dealt_cards[card], card);
-		add_card(b, *seatp, card);
+		if (from_seat)
+			remove_card(b, from_seat, card);
+		add_card(b, *to_seat, card);
 	}
+	b->par_score = -1;
+	invalidate_dd_scores (b);
 
 	board_statusbar(NULL);
 	card_window_update(b->dealt_cards);
-	show_board(b, REDRAW_HANDS);
+	show_board(b, REDRAW_HANDS | REDRAW_PAR);
 	PROTECT_END;
 }
 
+/*
 static void
 bidding_clicked (GtkTreeViewColumn *column, void *data)
 {
 	printf ("clicked bidd\n");
 }
+*/
 
+#if GTK_CHECK_VERSION (2,12,0)
 static gboolean
 bidding_query_tooltip (GtkWidget *widget, gint x, gint y, gboolean keyboard_mode,
 		GtkTooltip *tooltip, window_board_t *bidding_store)
@@ -449,6 +488,18 @@ bidding_query_tooltip (GtkWidget *widget, gint x, gint y, gboolean keyboard_mode
 	gtk_tooltip_set_markup (tooltip, !alert || *alert ? alert : _("(no explanation)"));
 	return TRUE;
 }
+#endif
+
+#if GTK_CHECK_VERSION (2,10,0)
+static void
+jump_menu_select (GtkWidget *recentchooser, char *unused)
+{
+	char *filename = gtk_recent_chooser_get_current_uri (GTK_RECENT_CHOOSER (recentchooser));
+	if (strncmp (filename, "file://", sizeof ("file://") - 1))
+		return;
+	board_load_popup (win, 0, filename + sizeof ("file://") - 1);
+}
+#endif
 
 /* infrastructure */
 
@@ -474,6 +525,7 @@ static void create_hand_widgets (window_board_t *win)
 	}
 	g_object_unref (sizegroup);
 
+	/* trick display */
 	GtkWidget *grid = lookup_widget (win->window, "table1");
 	GtkWidget *table = hand_display_new (HAND_DISPLAY_MODE_TABLE);
 	gtk_table_attach_defaults (GTK_TABLE (grid), table, 1, 2, 1, 2);
@@ -509,8 +561,10 @@ create_bidding_widget (window_board_t *win)
 		G_TYPE_STRING, G_TYPE_STRING,
 		G_TYPE_STRING, G_TYPE_STRING);
 	gtk_tree_view_set_model (win->bidding, GTK_TREE_MODEL (win->bidding_store));
+#if GTK_CHECK_VERSION (2,12,0)
 	g_signal_connect (G_OBJECT (win->bidding), "query-tooltip",
 			G_CALLBACK (bidding_query_tooltip), win);
+#endif
 
 	GtkCellRenderer *renderer;
 	renderer = gtk_cell_renderer_text_new ();
@@ -549,12 +603,13 @@ create_bidding_widget (window_board_t *win)
 
 void board_window_set_style (window_board_t *win, int style)
 {
+	win->hand_display_style = style;
 	int h;
 	for (h = 0; h < 4; h++) {
-		hand_display_set_style(win->handdisp[h], style, NULL);
+		hand_display_set_style(win->handdisp[h], style);
 	}
-	hand_display_set_style(win->table, style, NULL);
-	if (CUR_BOARD)
+	hand_display_set_style(win->table, style);
+	if (win->n_boards)
 		show_board(CUR_BOARD, REDRAW_HANDS);
 }
 
@@ -588,16 +643,33 @@ board_window_init (window_board_t *win)
 	create_hand_widgets(win);
 	create_bidding_widget (win);
 
+	/* set up "recently used" menu */
+	GtkWidget *jump_menu = lookup_widget(win->window, "jump_to1");
+#if GTK_CHECK_VERSION (2,10,0)
+	GtkWidget *recentchooser = gtk_recent_chooser_menu_new ();
+	gtk_recent_chooser_set_sort_type (GTK_RECENT_CHOOSER (recentchooser),
+			GTK_RECENT_SORT_MRU);
+	gtk_menu_item_set_submenu (GTK_MENU_ITEM (jump_menu), recentchooser);
+	g_signal_connect (G_OBJECT (recentchooser), "item-activated",
+			G_CALLBACK (jump_menu_select), NULL);
+	GtkRecentFilter *filter = gtk_recent_filter_new ();
+	gtk_recent_filter_add_pattern (filter, "*.lin");
+	gtk_recent_filter_add_pattern (filter, "*.pbn");
+	gtk_recent_chooser_add_filter (GTK_RECENT_CHOOSER (recentchooser), filter);
+#else
+	gtk_widget_set_sensitive (jump_menu, FALSE);
+#endif
+
+	win->keyfile = g_key_file_new ();
 	win->show_played_cards = 0;
 	win->show_hands = seat_all;
 	win->show_dd_scores = seat_all;
-	win->card_width = 80;
+	win->card_width = 70;
 
 	win->filename = NULL;
-	win->title = NULL;
-	win->boards = calloc(4, sizeof(board*));
-	assert (win->boards);
-	win->n_boards_alloc = 4;
+	win->title = win->subtitle = win->team1 = win->team2 = NULL;
+	win->boards = NULL;
+	win->n_boards_alloc = 0;
 	win->n_boards = 0;
 
 	win->cutbuf = NULL;
@@ -625,13 +697,23 @@ void board_statusbar (char *text)
 void
 board_set_declarer (seat declarer)
 {
-	PROTECT_BEGIN;
 	board *b = CUR_BOARD;
-	board_rewind(b);
+	if (declarer == b->declarer)
+		return;
+
+	if (b->n_played_cards > 0) {
+		board_statusbar (_("Cannot change declarer while cards are in play"));
+		PROTECT_BEGIN;
+		show_board(b, REDRAW_CONTRACT);
+		PROTECT_END;
+		return;
+	}
+
+	PROTECT_BEGIN;
 	b->declarer = declarer;
 	b->current_turn = seat_mod(declarer + 1);
+	invalidate_dd_scores (b);
 	show_board(b, REDRAW_CONTRACT | REDRAW_TRICKS | REDRAW_HANDS | REDRAW_NAMES | REDRAW_BOARD_LIST);
-	// FIXME: redraw less?
 	PROTECT_END;
 }
 
@@ -648,9 +730,22 @@ board_set_dealer (seat dealer)
 void
 board_set_trumps (suit trumps)
 {
-	PROTECT_BEGIN;
 	board *b = CUR_BOARD;
+	if (trumps == b->trumps)
+		return;
+
+	if (b->n_played_cards > 0) {
+		board_statusbar (_("Cannot change trumps while cards are in play"));
+		PROTECT_BEGIN;
+		show_board(b, REDRAW_CONTRACT);
+		PROTECT_END;
+		return;
+	}
+
+	PROTECT_BEGIN;
+	board_rewind (b);
 	b->trumps = trumps;
+	invalidate_dd_scores (b);
 	show_board(b, REDRAW_CONTRACT | REDRAW_BOARD_LIST);
 	PROTECT_END;
 }
@@ -658,10 +753,13 @@ board_set_trumps (suit trumps)
 void
 board_set_level (int level)
 {
-	PROTECT_BEGIN;
-	// FIXME: don't call show_board when called from show_board
 	board *b = CUR_BOARD;
+	if (level == b->level)
+		return;
+
+	PROTECT_BEGIN;
 	b->level = level;
+	invalidate_dd_scores (b);
 	calculate_target(b);
 	show_board(b, REDRAW_CONTRACT | REDRAW_BOARD_LIST);
 	PROTECT_END;
@@ -672,10 +770,12 @@ board_set_vuln (int ns, int ew)
 {
 	PROTECT_BEGIN;
 	board *b = CUR_BOARD;
+	if (ns == b->vuln[0] && ew == b->vuln[1])
+		PROTECT_RETURN;
 	b->vuln[0] = ns;
 	b->vuln[1] = ew;
 	b->par_score = -1;
-	show_board(b, REDRAW_CONTRACT | REDRAW_NAMES | REDRAW_BIDDING);
+	show_board(b, REDRAW_CONTRACT | REDRAW_NAMES | REDRAW_BIDDING | REDRAW_PAR);
 	PROTECT_END;
 }
 
@@ -688,49 +788,3 @@ board_set_doubled (int doubled)
 	show_board(b, REDRAW_CONTRACT | REDRAW_BOARD_LIST);
 	PROTECT_END;
 }
-
-int
-read_config (window_board_t *win)
-{
-	if (!rcfile) {
-		char *home = getenv ("HOME");
-		if (!home) {
-			printf ("HOME unset, cannot read config");
-			return 0;
-		}
-		rcfile = g_string_new (NULL);
-		g_string_printf (rcfile, "%s/.tenacerc", home);
-	}
-	FILE *f;
-	if (!(f = fopen(rcfile->str, "r"))) {
-		return 0;
-	}
-	int style;
-	if (fscanf (f, "style %d", &style) == 1) {
-		char *checkitem_name[] = { "style_text", "style_cards" };
-		GtkWidget *checkitem = lookup_widget (win->window, checkitem_name[style > 0]);
-		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (checkitem), TRUE);
-		board_window_set_style (win, style > 0 ? HAND_DISPLAY_STYLE_CARDS :
-				HAND_DISPLAY_STYLE_TEXT);
-	}
-	fclose (f);
-	return 1;
-}
-
-int
-write_config (window_board_t *win)
-{
-	if (!rcfile) {
-		return 0;
-	}
-	FILE *f;
-	if (!(f = fopen(rcfile->str, "w"))) {
-		perror (rcfile->str);
-		return 0;
-	}
-	GtkWidget *checkitem = lookup_widget (win->window, "style_cards");
-	fprintf (f, "style %d\n",
-		gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (checkitem)));
-	fclose (f);
-	return 1;
-}
