@@ -1,5 +1,6 @@
 void card_window_update (seat *cards);
 void set_new_card_seat (seat s);
-void window_card_init ();
+void window_card_set_style (int style);
+void window_card_init (int style);
 void window_card_delete (void);
 
